/*
  Copyright Red Hat, Inc. 2008

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; version 2.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
  MA 02139, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <wait.h>
#include "readlinkr.h"

/**
  Read a file from a stdout pipe.
 */
int
read_pipe(int fd, char **file, size_t *length)
{
	char buf[4096];
	int n, done = 0;

	*file = NULL;
	*length = 0;

	while (!done) {

		n = read(fd, buf, sizeof(buf));
		if (n < 0) {

			if (errno == EINTR)
				continue;

			if (*file)
				free(*file);
			return -1;
		}

		if (n == 0 && (!*length))
			return 0;

		if (n == 0) {
			done = 1;
		}

		if (*file)
			*file = realloc(*file, (*length) + n + done);
		else
			*file = malloc(n + done);

		if (!*file)
			return -1;

		memcpy((*file) + (*length), buf, n);
		*length += (done + n);
	}

	/* Null terminator */
	(*file)[(*length) - 1] = 0;

	return 0;
}


char *
capture_output(const char *command)
{
	int pid;
	int _pipe[2];
	char *data;
	size_t size;

	if (pipe(_pipe) == -1)
		return NULL;

	pid = fork();
	if (pid == -1) {
		close(_pipe[0]);
		close(_pipe[1]);
	}

	if (pid == 0) {
		/* child */
		close(0);
		close(1);
		close(2);

		close(_pipe[0]);
		dup2(_pipe[1], 1);
		close(_pipe[1]);
		
		/* exec */
		execl("/bin/sh", "/bin/sh", "-c", command, NULL);
		exit(1);
	}

	close(_pipe[1]);
	/* parent */
	if (read_pipe(_pipe[0], &data, &size) == -1) {
		close(_pipe[0]);
		return NULL;
	}

	waitpid(pid, NULL, 0);
	close(_pipe[0]);

	if (!size)
		return NULL;

	return data;
}


void
chomp(char *buf)
{
	int x;

	while (1) {
		x = strlen(buf);
		if (!x)
			break;
		if (buf[x-1] == '\n' || buf[x-1] == '\r') {
			buf[x-1] = 0;
			continue;
		}
		break;
	}
}


/**
 * TODO: Make it not fork findfs
 *
 * Given a real device name, find 
 */
int
find_fs(char *device, char *real_device, size_t real_dev_len)
{
	char buf[4096];
	char *ret_buf;

	if (device[0] != '/') {
		/* TODO - fork / pipe to findfs */
		snprintf(buf, sizeof(buf), "findfs %s\n", device);
		ret_buf = capture_output(buf);
		if (!ret_buf)
			return -1;
		chomp(ret_buf);
		if (strlen(ret_buf) == 0) {
			free(ret_buf);
			return -1;
		}
		strncpy(real_device, ret_buf, real_dev_len);
		free(ret_buf);
		return 0;
	}

	if (strlen(device) > (real_dev_len-1)) {
		errno = ENAMETOOLONG;
		return -1;
	}

	if (readlink_recurse(device, real_device, real_dev_len) < 0) {
		return -1;
	}

	return 0;
}
