/*
  Copyright Red Hat, Inc. 2008

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; version 2.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
  MA 02139, USA.
*/
#include <stdio.h>
#include <mntent.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mount.h>
#include <errno.h>
#include <string.h>
#include "device.h"
#include "metadata.h"
#include "fsc.h"


struct mount_flag {
	char *option;
	int bit;
};

/* These flags are done by normal mount(2).  The rest, we need 
  to pull out and provide to the file system */
/* These are from mount(2) */
static struct mount_flag mount_flags[] = {
	{	"bind",		MS_BIND },
	//{	"dirsync",	MS_DIRSYNC },
	{	"mandlock",	MS_MANDLOCK },
	//{	"move",		MS_MOVE },
	{	"noatime",	MS_NOATIME },
	{	"nodev",	MS_NODEV },
	{	"nodiratime",	MS_NODIRATIME },
	{	"diratime",	0 },
	{	"noexec",	MS_NOEXEC },
	{	"exec",		0 },
	{	"nosuid",	MS_NOSUID },
	{	"suid",		0 },
	{	"ro",		MS_RDONLY },
	{	"rw",		0 },
	{	"remount",	MS_REMOUNT },
	{	"sync",		MS_SYNCHRONOUS },
	{	"async",	0 },
	{	NULL, 0 }
};


/*
   take in mount options
   squeeze out global ones quickly
   return flags and string
 */
int
parse_mount_flags(char *mount_options_in, unsigned long *flags_ret, char *fs_opts, size_t fs_optsz)
{
	char mount_options[4096];
	char *start = mount_options;
	char *comma = NULL;
	unsigned long flags = 0;
	int x;
	int len, found;

	if (!mount_options_in || !strlen(mount_options_in)) {
		*flags_ret = flags;
		return 0;
	}

	if (strlen(mount_options_in) >= sizeof(mount_options)) {
		errno = ENAMETOOLONG;
		return -1;
	}

	snprintf(mount_options, sizeof(mount_options), "%s", mount_options_in);
	start = mount_options;

	do {
		comma = strchr(start, ',');
		if (comma) {
			*comma = 0;
			len = comma - start;
		} else {
			len = strlen(start);
		}
				
		found = 0;
		for (x = 0; mount_flags[x].option; x++) {
			if (!strcasecmp(mount_flags[x].option, start)) {
				/* got an option */
				flags |= mount_flags[x].bit;
				found = 1;
			}
		}

		if (found) {
			if (comma) {
				++comma;
				memmove(start, comma, strlen(comma)+1);
			} else {
				*start = 0;
				start = NULL;
			}
		} else if (comma) {
			*comma = ',';
			start = comma+1;
		} else {
			start = NULL;
		}
	} while (start);

	comma = &mount_options[strlen(mount_options)-1];
	if (*comma == ',')
		*comma = 0;

	*flags_ret = flags;
	snprintf(fs_opts, fs_optsz, "%s", mount_options);

	return 0;
}


/* 
   read in mount entries
   place entries in list structure
   return list structure
 */
mount_entries_t *
get_mount_list(const char *file)
{
	FILE *fp;
	struct mntent *me;
	mount_entries_t *ml;
	int count = 0;
	size_t totalsz;

	fp = setmntent(file, "r");
	
	while ((me = getmntent(fp))) {
		++count;
	}

	rewind(fp);

	totalsz = sizeof(*ml) + sizeof(struct mntent) * count;
	ml = malloc(totalsz);
	if (!ml)
		return NULL;
	memset(ml, 0, totalsz);

	ml->count = count;
	count = 0;
	while ((me = getmntent(fp)) && (count < ml->count)) {
		ml->entries[count].mnt_fsname = strdup(me->mnt_fsname);
		ml->entries[count].mnt_dir = strdup(me->mnt_dir);
		ml->entries[count].mnt_type = strdup(me->mnt_type);
		ml->entries[count].mnt_opts = strdup(me->mnt_opts);
		/* XXX todo - check return value of strdup */

		++count;
	}

	return ml;
}


/*
   free the memory
   occupied by a mount list
   for better or worse
 */
void
free_mount_list(mount_entries_t *ml)
{
	int x;
	
	if (!ml)
		return;

	for (x = 0; x < ml->count; x++) {
		if (ml->entries[x].mnt_fsname)
			free(ml->entries[x].mnt_fsname);
		if (ml->entries[x].mnt_dir)
			free(ml->entries[x].mnt_dir);
		if (ml->entries[x].mnt_type)
			free(ml->entries[x].mnt_type);
		if (ml->entries[x].mnt_opts)
			free(ml->entries[x].mnt_opts);
	}

	free(ml);
}


/*
   returns:
     0 - not mounted
     1 - mounted, and in the right place
     2 - mounted, and in the -wrong- place
 */
int
is_mounted(mount_entries_t *ml, char *device, char *location)
{
	int x;

	for (x = 0; x < ml->count; x++) {
		if (!strcmp(device, ml->entries[x].mnt_fsname)) {
			if (!location ||
			    !strcmp(location, ml->entries[x].mnt_dir))
				return 1;
			return 2;
		}
	}

	return 0;
}


#ifdef STANDALONE
int
main(int argc, char **argv)
{
	mount_entries_t *ml;
	int x;
	char dev[4096];
	if (argc == 2) {
		if (!strcmp(argv[1], "meta-data")) {
			xml_print_metadata(&fsc_agent);
			return 0;
		}
	}

	ml = get_mount_list("/proc/mounts");

	for (x = 0; x < ml->count; x++) {
		printf("--------------\n");
		printf("mnt_fsname = %s\n", ml->entries[x].mnt_fsname);
		printf("mnt_dir = %s\n", ml->entries[x].mnt_dir);
		printf("mnt_type = %s\n", ml->entries[x].mnt_type);
		printf("mnt_opts = %s\n", ml->entries[x].mnt_opts);
	}

	if (argc == 2) {
		printf("find_fs!\n");
		if (find_fs(argv[1], dev, sizeof(dev)) == 0) {
			printf("find_fs %s -> %s\n", argv[1], dev);
		}
	}

	if (argc == 3) {
		printf("is_mounted %s %s -> %d\n",
			argv[1], argv[2], is_mounted(ml, argv[1], argv[2]));
	}

	free_mount_list(ml);
	
	return 0;
}

#endif
